"""
用来管理不同的用户所创建的线程，以及线程的状态
"""
from typing import List, Dict
import os, sys
from pathlib import Path
here = Path(__file__).parent
import time
import json

try:
    from hepai.agents.version import __version__
except:
    sys.path.insert(1, f'{here.parent.parent.parent}')
    from hepai.agents.version import __version__
from hepai.agents.utils import BaseJsonSaver
from hepai.agents.version import __appname__
from hepai.agents.configs import CONST

from hepai.agents.modules.managers.base_thread_message import ThreadMessage, Content, Text
from hepai.agents.modules.managers.base_thread import Thread, ThreadDeleted
from hepai.agents.modules.managers.base_run import ThreadRun, TruncationStrategy
from hepai.agents.modules.managers.base_pages import CursorPage
from hepai.agents.utils import Logger, EventCollector

logger = Logger.get_logger("threads_manager.py")

class ThreadsManager(BaseJsonSaver):
    """
    离线存储所有的线程
    """

    version = "1.0.0"
    metadata = {
        "description": "Thread is conversation session between user and AI assistant, the ThreadsManages is used to store all threads",
        "mapping_username2indexes": {},  # 用来存储用户名到线程索引的映射，方便快速查找
        }

    def __init__(
        self,
        file_name: str = f'threads.json',
        file_dir: str = f'{Path.home()}/.{__appname__}',
        **kwargs
        ) -> None:
        super().__init__(auto_save=True, **kwargs)

        self.file_path = os.path.join(file_dir, file_name)
        self._data = self._init_load(self.file_path, version=self.version, metadata=self.metadata)

        self.message_save_dir = f"{file_dir}/thread_messages"
        self.run_save_dir = f"{file_dir}/thread_runs"
        if not os.path.exists(self.message_save_dir):
            os.makedirs(self.message_save_dir)
        if not os.path.exists(self.run_save_dir):
            os.makedirs(self.run_save_dir)

        self.alive_threads = dict()
        ## TODO: 新建的thread保存到内存中，方便快速查找，减少磁盘IO，thread被访问时，更新时间，超时的线程删除

    def save_messages_to_disk(self, file_path: str, messages: List[str | ThreadMessage]):
        """
        保存数据到硬盘，如果数据已经存在，不覆盖
        """
        # if os.path.exists(file_path):
        #     return
        if not messages:  # messages为空，不需要保存
            return
        if isinstance(messages[0], str):
            # 在一个线程上的一个message是str，表示其未被加载，不需要保存
            return
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, 'w') as f:
            for message in messages:
                msg_dict = message.to_dict(only_output_keys=True)
                f.write(json.dumps(msg_dict) + '\n')
        logger.debug(f"Messages saved to `{Path(file_path).relative_to(self.message_save_dir)}`")

    def save_runs_to_disk(self, file_path: str, run: List[str | ThreadRun]):
        # if os.path.exists(file_path):  # 如果文件已经存在，不覆盖，这个不对，一个线程的数据可能更新
        #     return
        if not run:  # run为空，不需要保存
            return
        if isinstance(run[0], str):
            # 在一个线程上的一个run是str，表示其未被加载，不需要保存
            return
        
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, 'w') as f:
            for run in run:
                run_dict = run.to_dict(only_output_keys=True)
                f.write(json.dumps(run_dict) + '\n')
        logger.debug(f"Runs saved to `{Path(file_path).relative_to(self.run_save_dir)}`")
        pass

    def save(self, file_path: str = None, data: dict = None, username=None, **kwargs):
        """
        单个thread的对象里可能存在runs和message对象
        无法直接序列化到json文件里，需要单独保存到硬盘中。
        Args:
            username: str, optional. The username of the thread owner.
        """
        data = data or self.data

        entities = data.get("entities", [])
        if username:
            entities = [x for x in entities if x.get("username") == username]

        for entity in entities:
            messages = entity.pop("messages", [])
            runs = entity.pop("runs", [])

            thread_id = entity["id"]
            username = entity["username"]

            # 单独保存message
            if len(messages) > 0:
                if isinstance(messages[0], str):
                    msg_ids = messages
                else:
                    msg_ids = [msg.id for msg in messages]
            else:
                msg_ids = []
            entity["messages"] = msg_ids
            msg_save_path = f"{self.message_save_dir}/{username}/messages_of_{thread_id}.jsonl"
            self.save_messages_to_disk(msg_save_path, messages)
        
            # 单独保存run
            if len(runs) > 0:
                if isinstance(runs[0], str):
                    run_ids = runs
                else:
                    run_ids = [run.id for run in runs]
            else:
                run_ids = []
            entity["runs"] = run_ids
            run_save_path = f"{self.run_save_dir}/{username}/runs_on_{thread_id}.jsonl"
            self.save_runs_to_disk(run_save_path, runs)

        return super().save(file_path=file_path, data=data)

    ### --- Threads --- ###
    def create_threads(self, username=None, **kwargs):
        """
        创建一个线程
        {
        "id": "thread_abc123",
        "object": "thread",
        "created_at": 1699012949,
        "metadata": {},
        "tool_resources": {}
        }
        """
        username = username or self.DEFAULT_USERNAME
        save_immediately = kwargs.get('save_immediately', False)
        messages: List[Dict] = kwargs.get('messages', [])
        tool_resources: List = kwargs.get('tool_resources', {})
        metadata: Dict = kwargs.get('metadata', {})

        # thread_id = self.auto_id(prefix='thread_', length=30)
        # 使用openwebui的chat_id作为thread_id
        chat_id = kwargs.get('chat_id', None)
        if chat_id:
            thread_id = 'thread_'+chat_id
        else:
            thread_id = self.auto_id(prefix='thread_', length=30)
            
        try:
        #    thread = self.alive_threads[thread_id]
        #    return thread
            return self.get_thread(thread_id, username=username)
        except:
            order_ids = [int(x['order_id']) for x in self.entities]
            order_id = int(max(order_ids)) if order_ids else 0
            order_id = f'{order_id+1:0>8}'

        metadata["Stream_status"] = "thread_start"

        thread = Thread(
            id=thread_id,
            object="thread",
            created_at=int(time.time()),
            metadata=metadata,
            tool_resources=tool_resources,
            username=username,
            order_id=order_id
        )

        if messages:
            raise NotImplementedError("Creating thread with messages is not supported yet.")
        self.append_entity(thread, username=username, save_immediately=save_immediately)
        self.alive_threads[thread_id]=thread
        return thread
        
    def load_messages_from_disk(self, file_path: str) -> List[ThreadMessage]:
        """
        从硬盘读取数据并转为List[ThreadMessage]
        """
        if not os.path.exists(file_path):
            return []  # 文件不存在，返回空列表

        messages = []
        with open(file_path, 'r') as f:
            for line in f:
                msg_dict = json.loads(line.strip())  # 将一行JSON字符串解析为字典
                message = ThreadMessage(**msg_dict)  # 根据字典创建ThreadMessage实例
                content = Content(
                    type="text",
                    text=Text(
                        value=msg_dict["content"][0]["text"]["value"], 
                        annotations=[]
                    ))
                message.content = [content]
                messages.append(message)
        
        logger.debug(f"Messages loaded from `{file_path}`")
        return messages
    
    def get_thread(self, thread_id, username=None, return_index=False):
        """从json文件获取线程"""
        if username:  # 提供用户名，缩小搜搜范围，搜索的快
            threads_indexes = self.data["metadata"]["mapping_username2indexes"].get(username, [])
            if not threads_indexes:
                raise ModuleNotFoundError(f"No thread found for user `{username}`")
        else:
            threads_indexes = list(range(len(self.entities)))
            if not threads_indexes:
                raise ModuleNotFoundError(f"No thread found")
        
        filtered = [(self.entities[idx], idx) for idx in threads_indexes if self.entities[idx]['id'] == thread_id]
        if not filtered:
            raise ModuleNotFoundError(f"No thread found for user `{username}` and id `{thread_id}`")
        threads, indexes = zip(*filtered)
        if not threads:
            raise ModuleNotFoundError(f"Thread not found, id: `{thread_id}`")
        if len(threads) > 1:
            raise ValueError(f"Multiple threads found, id: `{thread_id}`")
        thread = Thread(**threads[0])
        if len(thread.messages)>0:
            if isinstance(thread.messages[0],str):
                # 如果是从json文件中回复的，message=str，需要从硬盘加载message
                msg_save_path = f"{self.message_save_dir}/{username}/messages_of_{thread_id}.jsonl"
                messages = self.load_messages_from_disk(msg_save_path)
                thread.messages = messages
        if thread.deleted:
            raise ModuleNotFoundError(f"Thread has been deleted, id: `{thread_id}`")
        if return_index:
            assert len(indexes) == 1, f"Multiple threads found, id: `{thread_id}`"
            return thread, indexes[0]
        return thread

    def retrieve_thread(self, thread_id, username=None, **kwargs):
        """
        获取线程
        """
        return self.get_thread(thread_id, username=username, **kwargs)

    
    def update_thread(self, thread_id, username=None, **kwargs):
        save_immediately = kwargs.pop('save_immediately', False)
        thread, idx = self.get_thread(thread_id, username=username, return_index=True)

        # 还需自动更新是否修改的标识和谁更新的，存储到metadata中
        thread.update(auto_metadata=True, **kwargs)
        self.update_entity(thread, idx, save_immediately=save_immediately)
        return thread
    
    def delete_thread(self, thread_id, username=None, permanent=False, **kwargs):
        """
        Delete a thread
        Args:
            thread_id: str, required
            username: str, optional. Boost the search speed by providing the username.
            permanent: bool, optional. If True, the thread will be permanently deleted. Default is False.
        """
        save_immediately = kwargs.get('save_immediately', False)
        permanent = True if permanent == 'true' else False  # fit webui query
        if not permanent:  # 软删除，只标记为删除
            asst = self.update_thread(thread_id, username=username, deleted=True, save_immediately=save_immediately)
            assert asst.deleted, "Thread not deleted"
        else:
            asst, idx = self.get_thread(thread_id, username=username, return_index=True)
            self.remove_entity(asst, idx, save_immediately=save_immediately)
        return ThreadDeleted(id=thread_id, object="thread.deletec", deleted=True)
    
    ### --- Messages --- ###
    def get_message(self, thread_id, message_id, username=None):
        """Get message from thread"""
        thread = self.get_thread(thread_id, username=username)
        messages = thread.messages
        filtered = [msg for msg in messages if msg.id == message_id]
        if not filtered:
            raise ModuleNotFoundError(f"No message found for thread `{thread_id}` and id `{message_id}`")
        if len(filtered) > 1:
            raise ValueError(f"Multiple messages found, id: `{message_id}`")
        return filtered[0]
    
    def update_message(self, thread_id, message_id, username=None, **kwargs):
        """
        Update a message in a thread
        """
        message: ThreadMessage = self.get_message(thread_id, message_id, username=username)

        message.update(auto_metadata=True, **kwargs)

        # TODO: 保存到文件
        return message
        
    def retrieve_message(self, thread_id: str, message_id: str, username=None, **kwargs):
        """
        Retrieve a message from a thread
        """
        return self.get_message(thread_id, message_id, username=username, **kwargs)

    def list_messages(
            self,
            thread_id: str,
            limit: int = 20,
            order: str = 'desc',
            after: str = None,
            before: str = None, 
            run_id: str = None,
            username: str = None,
            **kwargs
        ) -> List[Dict]:
        """
        list messages of a thread
        :param thread_id: str, thread id, required
        :param limit: int, optinal. A limit on the number of objects to be returned. Limit can range between 1 and 100, and the default is 20.
        :param order: str, optinal. Sort order by the created_at timestamp of the objects. asc for ascending order and desc for descending order.
        :param after: str, optinal. A cursor for use in pagination. after is an object ID that defines your place in the list. For instance, if you make a list request and receive 100 objects, ending with obj_foo, your subsequent call can include after=obj_foo in order to fetch the next page of the list.
        :param before: str, optinal. A cursor for use in pagination. before is an object ID that defines your place in the list. For instance, if you make a list request and receive 100 objects, starting with obj_bar, your subsequent call can include before=obj_bar in order to fetch the previous page of the list.
        :praam run_id: str, optinal. Filter messages by the run ID that generated them.
        :param username: str, optinal. Boost the search speed by providing the username.
        """
        return_format = kwargs.get('return_format', Dict)

        thread = self.get_thread(thread_id, username=username)
        messages = thread.messages

        if run_id:  # 过滤run_id
            messages = [msg for msg in messages if msg.run_id == run_id]

        reverse = True if order == 'desc' else False
        messages = sorted(messages, key=lambda x: x.created_at, reverse=reverse)  # 默认升序
        if after:  # 其后的
            raise NotImplementedError("Pagination is not supported yet.")
            # cursor = [msg.id for msg in messages].index(after)
            # messages = [msg for msg in messages if msg.id > after]
        if before:  # 其前的
            raise NotImplementedError("Pagination is not supported yet.")
        has_more = len(messages) > limit
        if has_more:
            messages = messages[:limit]
        
        if len(messages) == 0:
            res = {
                "object": "list",
                "data": messages,
                "first_id": None,
                "last_id": None,
                "has_more": False,
            }
        else:
            res = {
                "object": "list",
                "data": messages,
                "first_id": messages[0].id,
                "last_id": messages[-1].id,
                "has_more": has_more,
            }
        if return_format in [Dict, dict, 'dict', 'Dict']:
            return res
        return CursorPage(**res)
    
   # 接收前端封装的消息，并打包成ThreadMessage，其中content需要进行格式判断
    def create_message(
            self, 
            thread: str | Thread, 
            role, 
            content, 
            **kwargs) -> ThreadMessage:
        """
        Create a message in a thread, auto append to the thread.
        Args:
            thread: str | Thread, required. Thread ID or Thread object.
            role: str, required. The role of the sender, can be 'user', 'assistant', 'system'.
            content: str, required. The content of the message.
            save_immediately: bool, optional. If True, save the message immediately. Default is False.
        """
        save_immediately = kwargs.get('save_immediately', False)
        stream = kwargs.get('stream', False)
        evc: EventCollector = kwargs.get('event_collector', None)
        username = kwargs.get('username', None)
        # NOTE: 这里不知道为啥非得限定'user', 'assistant', 'system'三个角色，请问是为了什么？这里进行了删除
        # assert role in ['user', 'assistant', 'system'], f"Invalid role: {role}"
        message_id = self.auto_id(prefix='msg_', length=30, deduplicate=False)

        if isinstance(thread, str):  
            thread: Thread = self.get_thread(thread, username=username)
        if thread.locked:
            raise ValueError(f"Thread `{thread.id}` is locked, cannot add new message.")
        
        logger.debug(f"Creating message in thread `{thread.id}` with role `{role}`")
        
        order_ids = [x.get("order_id") for x in thread.messages]
        order_id = int(max(order_ids))+1 if order_ids else 1
        order_id = f'{order_id:0>8}'

        if isinstance(content, str):
            content = [
                Content(
                    type="text",
                    text=Text(
                        value=content, 
                        annotations=kwargs.get('annotations', [])
                    )),
                ]
        elif isinstance(content, List):
            for i in range(len(content)):
                content[i]=Content(
                    type="text",
                    text=Text(
                        value=content[i].text.value, 
                        annotations=kwargs.get('annotations', [])
                    ))

        else:
            content = content

        thread_message = ThreadMessage(
            id=message_id,
            object="thread.message",
            created_at=int(time.time()),
            assistant_id=kwargs.get('assistant_id', None),
            thread_id=thread.id,
            run_id=kwargs.get('run_id', None),
            role=role,
            content=content,
            order_id=order_id,
            attachments=kwargs.get('attachments', []),
            metadata=kwargs.get('metadata', {}),
            sender=kwargs.get('sender', None),
            task_type=kwargs.get('task_type', None),
            )
        
        thread.messages.append(thread_message)
        thread_message.thread = thread  # 交叉引用

        if evc and stream:
            evc.add_event_source(thread_message.event_generator())
            thread_message.set_status("created", emit=True)

        return thread_message
    
    
        
    ### --- Runs --- ###
    async def a_create_runs(self, thread_id, assistant_id, **kwargs):
        """
        20240522，协程函数调用调试未通过
        """
        events_collector = kwargs.get('events_collector', None)
        run = await self.create_runs(thread_id, assistant_id, **kwargs)
        
        if events_collector:
            events_collector.append(self.run_created_event(run))
        return run

    def run_created_event(self, run):
        data = {
            "data": run.to_dict(),
            "event": "thread.run.created"
        }
        yield f"data: {json.dumps(data)}\n\n"

    def create_runs_stream(self, thread_id, assistant_id, **kwargs):
        thread_run = self.create_runs(thread_id, assistant_id, **kwargs)
        thread_run.set_status("created", emit=True)
        return thread_run, thread_run.event_generator()
        
    def create_runs(self, thread_id, assistant_id, **kwargs):
        """
        Create a run
        """
        save_immediately = kwargs.get('save_immediately', False)
        username = kwargs.get('username', None)

        thread: Thread = self.get_thread(thread_id, username=username)
        
        run_id = self.auto_id(prefix='run_', length=30, deduplicate=False)
        thread_run = ThreadRun(
            id=run_id,
            object="thread.run",
            created_at=int(time.time()),
            assistant_id=assistant_id,
            thread_id=thread.id,
            status="queued",
            started_at=None,
            expires_at=None,
            cancelled_at=None,
            failed_at=None,
            completed_at=None,
            last_error=None,
            model=kwargs.get('model', CONST.DEFAULT_MODEL),
            instructions=kwargs.get('instructions', None),
            incomplete_details=None,
            tools=kwargs.get('tools', []),
            metadata={},
            usage=None,
            temperature=kwargs.get('temperature', 0.6),
            top_p=kwargs.get('top_p', 1),
            max_prompt_tokens=1000,
            max_completion_tokens=1000,
            truncation_strategy=TruncationStrategy(type="auto", last_messages=None),
            response_format="auto",
            tool_choice="auto",
            username=thread.username,
        )

        # 交叉引用
        thread.runs.append(thread_run)  # 把run添加到线程中
        thread_run.thread = thread  # 把线程添加到run中
            
        return thread_run
    
    def get_run(self, thread_id, run_id, username=None):
        thread = self.get_thread(thread_id, username=username)
        runs = thread.runs
        filtered = [run for run in runs if run.id == run_id]
        if not filtered:
            raise ModuleNotFoundError(f"No run found for thread `{thread_id}` and id `{run_id}`")
        if len(filtered) > 1:
            raise ValueError(f"Multiple runs found, id: `{run_id}`")
        return filtered[0]

    def retrieve_run(self, thread_id, run_id, username=None, **kwargs):
        """
        Retrieve a run
        """
        return self.get_run(thread_id, run_id, username=username, **kwargs)
        

if __name__ == "__main__":
    tm = ThreadsManager()
    
    # 创建
    thread = tm.create_threads()
    print(f'Created thread: {thread}')

    # 获取
    my_thread = tm.retrieve_thread(thread.id)
    print(f'Retrieved thread: {my_thread}')
    assert my_thread.id == thread.id, "Thread ID not match"

    # 修改
    up_thread = tm.update_thread(thread.id, 
        metadata={'description': 'This is updated description'},
        )
    print(f'Updated thread: {up_thread}')
    assert up_thread.metadata['description'] == 'This is updated description', "Thread metadata not updated"
    assert up_thread.metadata['modified'], "Thread metadata not updated"

    # 删除
    # del_thread = tm.delete_thread(thread.id, permanent=True)
    # print(f'Deleted thread: {del_thread}')
    # assert del_thread.deleted, "Thread not deleted"

    message = tm.create_message(
        thread_id=thread.id, role='user', 
        content="How does AI work? Explain it in simple terms.")
    print(f"Created message: {message}")

    listed_messages = tm.list_messages(thread_id=thread.id)
    print(f"Listed messages: {listed_messages}")

    retrieved_message = tm.retrieve_message(thread_id=thread.id, message_id=message.id)
    print(f"Retrieved message: {retrieved_message}")

    up_message = tm.update_message(
        thread_id=thread.id, message_id=message.id, 
        content="How does AI work? Explain it in simple terms2.")
    print(f"Updated message: {up_message}")
    assert up_message.content == "How does AI work? Explain it in simple terms2.", "Message content not updated"
    assert up_message.metadata['modified'], "Message metadata not updated"

    #### Runs ####
    asst_id = "asst_39930f61c6574f139e1bca090"
    thread_run = tm.create_runs(thread_id=thread.id, assistant_id=asst_id)
    print(f"Created run: {thread_run}")

    pass



VALID_CHAR = set(['_', '-', '.', ':'])
PARAM_KEYS = ["data_id", "group"]
DEFAULT_GROUP_NAME = "DEFAULT_GROUP"


def is_valid(param):
    if not param:
        return False
    for i in param:
        if i.isalpha() or i.isdigit() or i in VALID_CHAR:
            continue
        return False
    return True


def check_params(params):
    for p in PARAM_KEYS:
        if p in params and not is_valid(params[p]):
            return False
    return True


def group_key(data_id, group, namespace):
    return "+".join([data_id, group, namespace])


def parse_key(key):
    sp = key.split("+")
    return sp[0], sp[1], sp[2]



import os

####################################
# Load .env file
####################################

try:
    from dotenv import load_dotenv, find_dotenv

    load_dotenv(find_dotenv("./.env"))
except ImportError:
    print("dotenv not installed, skipping...")

# 获取父目录
from pathlib import Path
here = Path(__file__).parent.absolute()

API_KEY = os.getenv("PIPELINES_API_KEY", "0p3n-w3bu!")
PIPELINES_DIR = os.getenv("PIPELINES_DIR", f"{here}/pipelines")





from typing import List
import os
import json
import uuid
from dataclasses import dataclass, field
import atexit
import schedule

import damei as dm
logger = dm.get_logger("base_json_saver.py")


class BaseJsonSaver:
    """
    提供一些基础操作本地某个特定Json文件的方法
    """
    file_path = None
    data = None
    version = "1.0.0"
    metadata = {}
    DEFAULT_USERNAME = "anonymous"


    def __init__(
        self, 
        auto_save: bool = False, 
        auto_save_interval: int = 10,  # 10 minutes
        debug: bool = False,
        **kwargs
        ) -> None:
        self._data = None
        self.dirty = False  # 是否有数据更新
        self.debug = debug

        if auto_save:
            atexit.register(self.check_and_save)
            # schedule.every(auto_save_interval).minutes.do(self.check_and_save)
            schedule.every(auto_save_interval).seconds.do(self.check_and_save)
            # schedule.run_all()

    @property
    def data(self):
        return self._data

    @property
    def metadata(self):
        return self._data["metadata"]
    
    @property
    def entities(self):
        return self._data.get("entities", {})
    
    @property
    def ids(self):
        ids = list(set([x.get('id', None) for x in self.entities]))
        return ids
        

    def _init_load(self, file_path: str, **kwargs):
        """
        从本地加载数据
        """
        if not os.path.exists(file_path):
            version = kwargs.get("version", self.version)
            metadata = kwargs.get("metadata", self.metadata)
            data = {
                "version": version,
                "metadata": metadata,
                "entities": []
            }
            # save
            if not os.path.exists(os.path.dirname(file_path)):
                os.makedirs(os.path.dirname(file_path))
            self.save(file_path, data)
            return data
        with open(file_path, 'r', encoding='utf-8') as f:
            data = json.load(f)
        return data

    def check_and_save(self, username=None, **kwargs):
        logger.debug("Check and save")
        if self.dirty:
            self.save(username=username, **kwargs)
    
    def save(self, file_path: str = None, data: dict = None, **kwargs):
        """
        save data to file
        """
        file_path = file_path or self.file_path
        data = data or self.data
        with open(file_path, 'w', encoding='utf-8') as f:
            json.dump(data, f, indent=4, ensure_ascii=False)
        logger.debug(f"Data saved to {file_path}")
        self.dirty = False
        return True
    
    def auto_id(self, prefix: str = '', length: int = 30, deduplicate: bool | List = True):
        """
        自动生成一个20位的id。
        """
        new_id = uuid.uuid4().hex
        short_id = str(new_id).replace('-', '')
        short_id = prefix + short_id[:length-len(prefix)]
        if deduplicate:
            if isinstance(deduplicate, list):
                if short_id in deduplicate:
                    return self.auto_id()
            else:
                if short_id in self.ids:  # 如果已经存在，则重新生成
                    return self.auto_id()
        return short_id
    
    def remove_entity(self, entity: dict | object, index: int, save_immediately: bool = False):
        """永久删除一个实体，并更新映射关系"""
        if not isinstance(entity, dict):
            entity = entity.to_dict()
        poped_entity = self.entities.pop(index)
        assert poped_entity.get('id') == entity.get('id'), "poped_entity.id must equal to entity.id"
       
        # 更新映射关系
        new_mapping = dict()
        for i, ent in enumerate(self.entities):
            username = ent.get('username', self.DEFAULT_USERNAME)
            if username not in new_mapping:
                new_mapping[username] = [i]
            else:
                new_mapping[username].append(i)
        self.data["metadata"]["mapping_username2indexes"] = new_mapping
            
        if save_immediately:
            self.save()
            return
        self.dirty = True
    
    def update_entity(self, entity: dict | object, index: int, save_immediately: bool = False):
        """
        更新一个实体
        """
        if not isinstance(entity, dict):
            entity = entity.to_dict()
        assert entity.get("id", None) is not None, "entity must have an id"
        self.entities[index] = entity
        if save_immediately:
            self.save()
            return
        self.dirty = True


    def append_entity(self, entity: dict | object, username=None, save_immediately: bool = False):
        """
        添加一个实体
        """
        if not isinstance(entity, dict):
            entity = entity.to_dict()
        assert entity.get("id", None) is not None, "entity must have an id"
        assert entity.get('id') not in self.ids, "id must be unique"
        self.entities.append(entity)

        # 保存映射关系，便于快速查找
        username = username or self.DEFAULT_USERNAME
        current_index = len(self.entities) - 1
        old_indexes =self.data["metadata"]["mapping_username2indexes"].get(username, [])
        assert current_index not in old_indexes, "current_index must not in old_indexes"
        new_indexes = old_indexes + [current_index]
        self.data['metadata']['mapping_username2indexes'][username] = new_indexes
        
        if save_immediately:
            self.save()
            return
        self.dirty = True  # 标记为脏数据，以便自动保存







import httpx
# from .openai_api.adapted_openai._exceptions import APIStatusError
from .openai_api import APIStatusError




class HAPIStatusError(APIStatusError):
    response: httpx.Response
    status_code: int
    request_id: str | None

    def __init__(self, message: str, *, response: httpx.Response, body: object | None) -> None:
        super().__init__(message, response=response, body=body)




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .jobs import (
    Jobs,
    AsyncJobs,
    JobsWithRawResponse,
    AsyncJobsWithRawResponse,
    JobsWithStreamingResponse,
    AsyncJobsWithStreamingResponse,
)
from ..._compat import cached_property
from .jobs.jobs import Jobs, AsyncJobs
from ..._resource import SyncAPIResource, AsyncAPIResource

__all__ = ["FineTuning", "AsyncFineTuning"]


class FineTuning(SyncAPIResource):
    @cached_property
    def jobs(self) -> Jobs:
        return Jobs(self._client)

    @cached_property
    def with_raw_response(self) -> FineTuningWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return the
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return FineTuningWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> FineTuningWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return FineTuningWithStreamingResponse(self)


class AsyncFineTuning(AsyncAPIResource):
    @cached_property
    def jobs(self) -> AsyncJobs:
        return AsyncJobs(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncFineTuningWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return the
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return AsyncFineTuningWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncFineTuningWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return AsyncFineTuningWithStreamingResponse(self)


class FineTuningWithRawResponse:
    def __init__(self, fine_tuning: FineTuning) -> None:
        self._fine_tuning = fine_tuning

    @cached_property
    def jobs(self) -> JobsWithRawResponse:
        return JobsWithRawResponse(self._fine_tuning.jobs)


class AsyncFineTuningWithRawResponse:
    def __init__(self, fine_tuning: AsyncFineTuning) -> None:
        self._fine_tuning = fine_tuning

    @cached_property
    def jobs(self) -> AsyncJobsWithRawResponse:
        return AsyncJobsWithRawResponse(self._fine_tuning.jobs)


class FineTuningWithStreamingResponse:
    def __init__(self, fine_tuning: FineTuning) -> None:
        self._fine_tuning = fine_tuning

    @cached_property
    def jobs(self) -> JobsWithStreamingResponse:
        return JobsWithStreamingResponse(self._fine_tuning.jobs)


class AsyncFineTuningWithStreamingResponse:
    def __init__(self, fine_tuning: AsyncFineTuning) -> None:
        self._fine_tuning = fine_tuning

    @cached_property
    def jobs(self) -> AsyncJobsWithStreamingResponse:
        return AsyncJobsWithStreamingResponse(self._fine_tuning.jobs)




__version__ = "0.5.6"
__appname__ = 'DrSai'
__author__ = 'hepai'
__email__ = 'hepai@ihep.ac.cn'
__description__ = 'A development framework for single and multi-agent collaborative systems developed by the Dr.Sai team at the IHEP, CAS.'


from typing import (
    AsyncGenerator, List, Sequence, 
    Dict, Any, Callable, Awaitable, 
    Union, Optional, Tuple)

import asyncio
import logging
import inspect
import json
import os

from pydantic import BaseModel

from autogen_core import CancellationToken, FunctionCall
from autogen_core.tools import BaseTool, FunctionTool, StaticWorkbench, Workbench, ToolResult, TextResultContent
from autogen_core.memory import Memory
from autogen_core.model_context import ChatCompletionContext
from autogen_core.models import (
    ChatCompletionClient,
    CreateResult,
    FunctionExecutionResultMessage,
    FunctionExecutionResult,
    LLMMessage,
    UserMessage,
    AssistantMessage,
    SystemMessage,
    RequestUsage,
)
from autogen_core import EVENT_LOGGER_NAME
from autogen_agentchat.agents import AssistantAgent
from autogen_agentchat.base import Handoff as HandoffBase
from autogen_agentchat.base import Response
from autogen_agentchat.messages import (
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MemoryQueryEvent,
    ModelClientStreamingChunkEvent,
    TextMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
    UserInputRequestedEvent,
    ThoughtEvent
)

from hepai.agents import HepAIChatCompletionClient
from hepai.agents.modules.managers.base_thread import Thread
from hepai.agents.modules.managers.threads_manager import ThreadsManager
from hepai.agents.modules.managers.base_thread_message import ThreadMessage, Content, Text

event_logger = logging.getLogger(EVENT_LOGGER_NAME)


class TextContent(BaseModel):
    type: str
    text: str

class DrSaiAgent(AssistantAgent):
    """基于aotogen AssistantAgent的定制Agent"""
    def __init__(
        self,
        name: str,
        *,
        model_client: ChatCompletionClient = None,
        tools: List[BaseTool[Any, Any] | Callable[..., Any] | Callable[..., Awaitable[Any]]] | None = None,
        workbench: Workbench | None = None,
        handoffs: List[HandoffBase | str] | None = None,
        model_context: ChatCompletionContext | None = None,
        description: str = "An agent that provides assistance with ability to use tools.",
        system_message: (
            str | None
        ) = "You are a helpful AI assistant. Solve tasks using your tools. Reply with TERMINATE when the task has been completed.",
        model_client_stream: bool = True,
        reflect_on_tool_use: bool | None = None,
        tool_call_summary_format: str = "{result}",
        output_content_type: type[BaseModel] | None = None,
        output_content_type_format: str | None = None,
        memory: Sequence[Memory] | None = None,
        metadata: Dict[str, str] | None = None,
        memory_function: Callable = None,
        reply_function: Callable = None,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs,
            ):
        '''
        memory_function: 自定义的memory_function，用于RAG检索等功能，为大模型回复增加最新的知识
        reply_function: 自定义的reply_function，用于自定义对话回复的定制
        '''
        if not model_client:
            model_client = HepAIChatCompletionClient(model="openai/gpt-4o", api_key=os.environ.get("HEPAI_API_KEY"))
        
        super().__init__(
            name, 
            model_client,
            tools=tools,
            workbench=workbench,
            handoffs=handoffs,
            model_context=model_context,
            description=description,
            system_message=system_message,
            model_client_stream=model_client_stream,
            reflect_on_tool_use=reflect_on_tool_use,
            tool_call_summary_format=tool_call_summary_format,
            output_content_type=output_content_type,
            output_content_type_format=output_content_type_format,
            memory=memory,
            metadata=metadata
            )
        
        self._reply_function: Callable = reply_function
        self._memory_function: Callable = memory_function
        self._thread: Thread = thread
        self._thread_mgr: ThreadsManager = thread_mgr
        self._user_params: Dict[str, Any] = {}
        self._user_params.update(kwargs)

    async def llm_messages2oai_messages(self, llm_messages: List[LLMMessage]) -> List[Dict[str, str]]:
        """Convert a list of LLM messages to a list of OAI chat messages."""
        messages = []
        for llm_message in llm_messages:
            if isinstance(llm_message, SystemMessage):
                messages.append({"role": "system", "content": llm_message.content} )
            if isinstance(llm_message, UserMessage):
                messages.append({"role": "user", "content": llm_message.content, "name": llm_message.source})
            if isinstance(llm_message, AssistantMessage):
                messages.append({"role": "assistant", "content": llm_message.content, "name": llm_message.source})
            if isinstance(llm_message, FunctionExecutionResultMessage):
                messages.append({"role": "function", "content": llm_message.content})
        return messages
    
    async def oai_messages2llm_messages(self, oai_messages: List[Dict[str, str]]) -> List[LLMMessage]:
        """Convert a list of OAI chat messages to a list of LLM messages."""
        messages = []
        for oai_message in oai_messages:
            if oai_message["role"] == "system":
                messages.append(SystemMessage(content=oai_message["content"]))
            if oai_message["role"] == "user":
                messages.append(UserMessage(content=oai_message["content"], source=oai_message.get("name", self.name)))
            if oai_message["role"] == "assistant":
                messages.append(AssistantMessage(content=oai_message["content"], source=oai_message.get("name", self.name)))
            if oai_message["role"] == "function":
                messages.append(FunctionExecutionResultMessage(content=oai_message["content"]))
        return messages
    
    async def _call_memory_function(self, llm_messages: List[LLMMessage]):
        """使用自定义的memory_function，为大模型回复增加最新的知识"""
        # memory_function: 自定义的memory_function，用于RAG检索等功能，为大模型回复增加最新的知识
        memory_messages = await self.llm_messages2oai_messages(llm_messages)
        try:
            memory_messages_with_new_knowledge: List[Dict[str, str]] = await self._memory_function(memory_messages, **self._user_params)
            llm_messages = await self.oai_messages2llm_messages(memory_messages_with_new_knowledge)
            return llm_messages
        except Exception as e:
            raise ValueError(f"Error: memory_function: {self._memory_function.__name__} failed with error {e}.")
    
    async def _call_reply_function(
            self, 
            llm_messages: List[LLMMessage],
            model_client: ChatCompletionClient,
            tools: Workbench,
            agent_name: str,
            cancellation_token: CancellationToken,
            thread: Thread = None,
            thread_mgr: ThreadsManager = None,
            **kwargs,
            ):
        """使用自定义的reply_function，自定义对话回复的定制"""

        oai_messages = await self.llm_messages2oai_messages(llm_messages)

        model_result: Optional[CreateResult] = None
        allowed_events = [
            ToolCallRequestEvent,
            ToolCallExecutionEvent,
            MemoryQueryEvent,
            UserInputRequestedEvent,
            ModelClientStreamingChunkEvent,
            ThoughtEvent]
        
        if self._model_client_stream:
            # 如果reply_function不是返回一个异步生成器而使用了流式模式，则会报错
            if not inspect.isasyncgenfunction(self._reply_function):
                raise ValueError("reply_function must be AsyncGenerator function if model_client_stream is True.")
            # Stream the reply_function.
            response = ""
            async for chunk in self._reply_function(
                oai_messages, 
                agent_name = agent_name,
                llm_messages = llm_messages, 
                model_client=model_client, 
                tools=tools, 
                cancellation_token=cancellation_token, 
                thread=thread, 
                thread_mgr=thread_mgr, 
                **self._user_params
                ):
                if isinstance(chunk, str):
                    response += chunk
                    yield ModelClientStreamingChunkEvent(content=chunk, source=agent_name)
                elif any(isinstance(chunk, event_type) for event_type in allowed_events):
                    response += str(chunk.content)
                    yield chunk
                elif isinstance(chunk, CreateResult):
                    model_result = chunk
                else:
                    raise RuntimeError(f"Invalid chunk type: {type(chunk)}")
            if isinstance(model_result, CreateResult):
                pass
            elif model_result is None:
            #     if isinstance(chunk, str):
            #         yield ModelClientStreamingChunkEvent(content=chunk, source=agent_name)
            #         response += chunk
            #     elif isinstance(chunk, AgentEvent):
            #         yield chunk
            #     elif isinstance(chunk, BaseAgentEvent):
            #     else:
            #         raise RuntimeError(f"Invalid chunk type: {type(chunk)}")
                assert isinstance(response, str)
                model_result = CreateResult(
                    content=response, finish_reason="stop",
                    usage = RequestUsage(prompt_tokens=0, completion_tokens=0),
                    cached=False)
        else:
            # 如果reply_function不是异步函数，或者是一个异步生成器，则会报错
            if not asyncio.iscoroutinefunction(self._reply_function) and not inspect.isasyncgenfunction(self._reply_function):
                raise ValueError("reply_function must be a coroutine function if model_client_stream is False.")
            response = await self._reply_function(
                oai_messages, 
                agent_name = agent_name,
                llm_messages = llm_messages, 
                tools=tools, 
                cancellation_token=cancellation_token, 
                **self._user_params
                )
            if isinstance(response, str):
                model_result = CreateResult(
                    content=response, finish_reason="stop",
                    usage = RequestUsage(prompt_tokens=0, completion_tokens=0),
                    cached=False)
            elif isinstance(response, CreateResult):
                model_result = response
            else:
                raise RuntimeError(f"Invalid response type: {type(response)}")
        yield model_result

    @classmethod
    async def call_llm(
        cls,
        agent_name: str,
        model_client: ChatCompletionClient,
        llm_messages: List[LLMMessage], 
        tools: List[BaseTool[Any, Any]], 
        model_client_stream: bool,
        cancellation_token: CancellationToken,
        output_content_type: type[BaseModel] | None,
        ) -> AsyncGenerator:
    
        model_result: Optional[CreateResult] = None

        if model_client_stream:
                
            async for chunk in model_client.create_stream(
                llm_messages, 
                tools=tools,
                json_output=output_content_type,
                cancellation_token=cancellation_token
            ):
                if isinstance(chunk, CreateResult):
                    model_result = chunk
                elif isinstance(chunk, str):
                    yield ModelClientStreamingChunkEvent(content=chunk, source=agent_name)
                else:
                    raise RuntimeError(f"Invalid chunk type: {type(chunk)}")
            if model_result is None:
                raise RuntimeError("No final model result in streaming mode.")
            yield model_result
        else:
            model_result = await model_client.create(
                llm_messages, tools=tools, cancellation_token=cancellation_token
            )
            yield model_result


### autogen 更改源码区

    async def _call_llm(
        self,
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        system_messages: List[SystemMessage],
        model_context: ChatCompletionContext,
        workbench: Workbench,
        handoff_tools: List[BaseTool[Any, Any]],
        agent_name: str,
        cancellation_token: CancellationToken,
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[Union[CreateResult, ModelClientStreamingChunkEvent], None]:
        """
        Perform a model inference and yield either streaming chunk events or the final CreateResult.
        """
        all_messages = await model_context.get_messages()
        if self._thread is None:
            pass
        else:
            # 从thread中获取历史消息，并于autogen中的消息记录合并
            history_aoi_messages: List[Dict[str, str]] = self._thread.metadata["history_aoi_messages"]
            history = []
            for  history_aoi_message in history_aoi_messages:
                if  history_aoi_message["role"] == "user":
                    history.append(UserMessage(content=history_aoi_message["content"], source=history_aoi_message["name"]))
                elif history_aoi_message["role"] == "assistant":
                    history.append(UserMessage(content=history_aoi_message["content"], source=history_aoi_message["name"]))
                elif history_aoi_message["role"] == "system":
                    history.append(SystemMessage(content=history_aoi_message["content"]))
                elif history_aoi_message["role"] == "function":
                    history.append(FunctionExecutionResultMessage(content=history_aoi_message["content"]))
            all_messages = history + all_messages
        
        llm_messages: List[LLMMessage] = self._get_compatible_context(model_client=model_client, messages=system_messages + all_messages)

        # 自定义的memory_function，用于RAG检索等功能，为大模型回复增加最新的知识
        if self._memory_function is not None:
            llm_messages = await self._call_memory_function(llm_messages)

        tools = (await workbench.list_tools()) + handoff_tools
        all_tools = tools + handoff_tools
        # model_result: Optional[CreateResult] = None
        if self._reply_function is not None:
            # 自定义的reply_function，用于自定义对话回复的定制
            async for chunk in self._call_reply_function(
                llm_messages, 
                model_client = model_client, 
                tools=workbench, 
                agent_name=agent_name, 
                cancellation_token=cancellation_token,
                thread = self._thread,
                thread_mgr = self._thread_mgr,
            ):
                # if isinstance(chunk, CreateResult):
                #     model_result = chunk
                yield chunk
        else:
           async for chunk in self.call_llm(
                agent_name = agent_name,
                model_client = model_client,
                llm_messages = llm_messages, 
                tools = all_tools, 
                model_client_stream = model_client_stream,
                cancellation_token = cancellation_token,
                output_content_type = output_content_type,
           ):
            #    if isinstance(chunk, CreateResult):
            #         model_result = chunk
               yield chunk


from ._completions import (
    ResponseFormatT as ResponseFormatT,
    has_parseable_input,
    has_parseable_input as has_parseable_input,
    maybe_parse_content as maybe_parse_content,
    validate_input_tools as validate_input_tools,
    parse_chat_completion as parse_chat_completion,
    get_input_tool_by_name as get_input_tool_by_name,
    solve_response_format_t as solve_response_format_t,
    parse_function_tool_arguments as parse_function_tool_arguments,
    type_to_response_format_param as type_to_response_format_param,
)
